"""Search class implementation."""

import logging
import os
from concurrent.futures import ThreadPoolExecutor, wait
from math import ceil
from random import random, shuffle
from types import TracebackType
from typing import Any, ClassVar

from .base import BaseSearchEngine
from .engines import ENGINES
from .exceptions import SearchException, TimeoutException
from .results import ResultsAggregator
from .similarity import SimpleFilterRanker
from .utils import _expand_proxy_tb_alias

logger = logging.getLogger(__name__)


class Search:
    """Seekify - A metasearch library for simple and distributed search.

    A metasearch library that aggregates results from diverse web search services.

    Args:
        proxy: The proxy to use for the search. Defaults to None.
        timeout: The timeout for the search. Defaults to 5.
        verify: bool (True to verify, False to skip) or str path to a PEM file. Defaults to True.

    Attributes:
        threads: The number of threads to use for the search. Defaults to None (automatic).
        _executor: The ThreadPoolExecutor instance.

    Raises:
        SearchException: If an error occurs during the search.

    Example:
        >>> from seekify import Search
        >>> results = Search().search("python")

    """

    threads: ClassVar[int | None] = None
    _executor: ClassVar[ThreadPoolExecutor | None] = None

    def __init__(self, proxy: str | None = None, timeout: int | None = 5, *, verify: bool | str = True) -> None:
        self._proxy = _expand_proxy_tb_alias(proxy) or os.environ.get("SEEKIFY_PROXY")
        self._timeout = timeout
        self._verify = verify
        self._engines_cache: dict[
            type[BaseSearchEngine[Any]], BaseSearchEngine[Any]
        ] = {}  # dict[engine_class, engine_instance]

    def __enter__(self) -> "Search":
        """Enter the context manager and return the Search instance."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        """Exit the context manager."""

    @classmethod
    def get_executor(cls) -> ThreadPoolExecutor:
        """Get a ThreadPoolExecutor instance and cache it."""
        if cls._executor is None:
            cls._executor = ThreadPoolExecutor(max_workers=cls.threads, thread_name_prefix="Seekify")
        return cls._executor

    def _get_engines(
        self,
        category: str,
        backend: str,
    ) -> list[BaseSearchEngine[Any]]:
        """Retrieve a list of search engine instances for a given category and backend.

        Args:
            category: The category of search engines (e.g., 'text', 'images', etc.).
            backend: A single or comma-delimited backends. Defaults to "auto".

        Returns:
            A list of initialized search engine instances corresponding to the specified
            category and backend. Instances are cached for reuse.

        """
        if isinstance(backend, list):  # deprecated handling
            backend = ",".join(backend)
        backend_list = [x.strip() for x in backend.split(",")]
        
        # Get all available engine keys for the category
        engine_keys = list(ENGINES[category].keys())
        shuffle(engine_keys)

        # Determine which keys to use
        if "auto" in backend_list or "all" in backend_list:
            keys = engine_keys
            # For text search, prioritize Wikipedia/Grokipedia if we are falling back to auto/all?
            # Actually the original logic seemed to force them to the front.
            if category == "text":
                priority_keys = ["wikipedia", "grokipedia"]
                other_keys = [k for k in keys if k not in priority_keys]
                keys = priority_keys + other_keys
        else:
            keys = backend_list

        try:
            # Instantiate engines or retrieve from cache
            available_engines = []
            for key in keys:
                engine_class = ENGINES[category].get(key)
                if not engine_class:
                    raise KeyError(key)

                if engine_class in self._engines_cache:
                    available_engines.append(self._engines_cache[engine_class])
                else:
                    engine_instance = engine_class(
                        proxy=self._proxy, 
                        timeout=self._timeout, 
                        verify=self._verify
                    )
                    self._engines_cache[engine_class] = engine_instance
                    available_engines.append(engine_instance)

            # Sort engines by priority (higher priority first)
            available_engines.sort(key=lambda e: (e.priority, random), reverse=True)
            return available_engines
        except KeyError as ex:
            logger.warning(
                "%r - backend is not exist or disabled. Available: %s. Using 'auto'",
                ex,
                ", ".join(sorted(engine_keys)),
            )
            return self._get_engines(category, "auto")

    def _search(  # noqa: C901
        self,
        category: str,
        query: str,
        keywords: str | None = None,  # deprecated
        *,
        region: str = "us-en",
        safesearch: str = "moderate",
        timelimit: str | None = None,
        max_results: int | None = 10,
        page: int = 1,
        backend: str = "auto",
        **kwargs: str,
    ) -> list[dict[str, Any]]:
        """Perform a search across engines in the given category.

        Args:
            category: The category of search engines (e.g., 'text', 'images', etc.).
            query: The search query.
            keywords: Deprecated alias for `query`.
            region: The region to use for the search (e.g., us-en, uk-en, ru-ru, etc.).
            safesearch: The safesearch setting (e.g., on, moderate, off).
            timelimit: The timelimit for the search (e.g., d, w, m, y) or custom date range.
            max_results: The maximum number of results to return. Defaults to 10.
            page: The page of results to return. Defaults to 1.
            backend: A single or comma-delimited backends. Defaults to "auto".
            **kwargs: Additional keyword arguments to pass to the search engines.

        Returns:
            A list of dictionaries containing the search results.

        """
        query = keywords or query
        if not query:
            msg = "query is mandatory."
            raise SearchException(msg)

        engines = self._get_engines(category, backend)
        # Unique providers to prevent duplicate queries to the same source (e.g. Bing)
        len_unique_providers = len({engine.provider for engine in engines})
        seen_providers: set[str] = set()

        # Perform search
        # Initialize results and executor
        results_aggregator: ResultsAggregator[set[str]] = ResultsAggregator({"href", "image", "url", "embed_url"})
        
        # Determine strict number of workers to prevent over-fetching
        max_workers = min(len_unique_providers, ceil(max_results / 10) + 1) if max_results else len_unique_providers
        
        executor = self.get_executor()
        futures, err = {}, None
        for i, engine in enumerate(engines, start=1):
            if engine.provider in seen_providers:
                continue
            future = executor.submit(
                engine.search,
                query,
                region=region,
                safesearch=safesearch,
                timelimit=timelimit,
                page=page,
                **kwargs,
            )
            futures[future] = engine

            if len(futures) >= max_workers or i >= max_workers:
                done, not_done = wait(futures, timeout=self._timeout, return_when="FIRST_EXCEPTION")
                for f, f_engine in futures.items():
                    if f in done:
                        try:
                            if r := f.result():
                                results_aggregator.extend(r)
                                seen_providers.add(f_engine.provider)
                        except Exception as ex:  # noqa: BLE001
                            err = ex
                            logger.info("Error in engine %s: %r", engine.name, ex)
                futures = {f: futures[f] for f in not_done}

            # Check if we have enough results to stop
            if max_results and len(results_aggregator) >= max_results:
                break

        results = results_aggregator.extract_dicts()
        # Rank results
        ranker = SimpleFilterRanker()
        results = ranker.rank(results, query)

        if results:
            return results[:max_results] if max_results else results

        if "timed out" in f"{err}":
            raise TimeoutException(err)
        raise SearchException(err or "No results found.")

    def text(self, query: str, **kwargs: Any) -> list[dict[str, Any]]:  # noqa: ANN401
        """Perform a text search."""
        return self._search("text", query, **kwargs)

    def images(self, query: str, **kwargs: Any) -> list[dict[str, Any]]:  # noqa: ANN401
        """Perform an image search."""
        return self._search("images", query, **kwargs)

    def news(self, query: str, **kwargs: Any) -> list[dict[str, Any]]:  # noqa: ANN401
        """Perform a news search."""
        return self._search("news", query, **kwargs)

    def videos(self, query: str, **kwargs: Any) -> list[dict[str, Any]]:  # noqa: ANN401
        """Perform a video search."""
        return self._search("videos", query, **kwargs)

    def books(self, query: str, **kwargs: Any) -> list[dict[str, Any]]:  # noqa: ANN401
        """Perform a book search."""
        return self._search("books", query, **kwargs)



"""Base class for search engines."""

import logging
from abc import ABC, abstractmethod
from collections.abc import Mapping
from functools import cached_property
from typing import Any, ClassVar, Generic, Literal, TypeVar

from lxml import html
from lxml.etree import HTMLParser as LHTMLParser

from .client import Client
from .results import BooksResult, ImagesResult, NewsResult, TextResult, VideosResult

logger = logging.getLogger(__name__)
T = TypeVar("T")


class BaseSearchEngine(ABC, Generic[T]):
    """Abstract base class for all search-engine backends."""

    name: ClassVar[str]  # unique key, e.g. "google"
    category: ClassVar[Literal["text", "images", "videos", "news", "books"]]
    provider: ClassVar[str]  # source of the search results (e.g. "bing" for DuckDuckgo)
    disabled: ClassVar[bool] = False  # if True, the engine is disabled
    priority: ClassVar[float] = 1

    search_url: str
    search_method: ClassVar[str]  # GET or POST
    search_headers: ClassVar[Mapping[str, str]] = {}
    items_xpath: ClassVar[str]
    elements_xpath: ClassVar[Mapping[str, str]]
    elements_replace: ClassVar[Mapping[str, str]]

    def __init__(self, proxy: str | None = None, timeout: int | None = None, *, verify: bool | str = True) -> None:
        self.client = Client(proxy=proxy, timeout=timeout, verify=verify)
        self.results: list[T] = []

    @property
    def result_type(self) -> type[T]:
        """Get result type based on category."""
        categories = {
            "text": TextResult,
            "images": ImagesResult,
            "videos": VideosResult,
            "news": NewsResult,
            "books": BooksResult,
        }
        return categories[self.category]

    @abstractmethod
    def build_payload(
        self,
        query: str,
        region: str,
        safesearch: str,
        timelimit: str | None,
        page: int,
        **kwargs: str,
    ) -> dict[str, Any]:
        """Build a payload for the search request."""
        raise NotImplementedError

    def request(self, *args: Any, **kwargs: Any) -> str | None:  # noqa: ANN401
        """Make a request to the search engine."""
        resp = self.client.request(*args, **kwargs)
        if resp.status_code == 200:
            return resp.text
        return None

    @cached_property
    def parser(self) -> LHTMLParser:
        """Get HTML parser."""
        return LHTMLParser(remove_blank_text=True, remove_comments=True, remove_pis=True, collect_ids=False)

    def extract_tree(self, html_text: str) -> html.Element:
        """Extract html tree from html text."""
        return html.fromstring(html_text, parser=self.parser)

    def pre_process_html(self, html_text: str) -> str:
        """Pre-process html_text before extracting results."""
        return html_text

    def extract_results(self, html_text: str) -> list[T]:
        """Extract search results from html text.

        1. Pre-process the HTML.
        2. Parse the HTML into a tree.
        3. Find all items using the `items_xpath`.
        4. For each item, extract fields using `elements_xpath` and populate the result object.
        """
        html_text = self.pre_process_html(html_text)
        tree = self.extract_tree(html_text)
        items = tree.xpath(self.items_xpath)
        results = []
        for item in items:
            result = self.result_type()
            for key, value in self.elements_xpath.items():
                # Extract text content from the XPath match
                data = " ".join(x.strip() for x in item.xpath(value))
                result.__setattr__(key, data)
            results.append(result)
        return results

    def post_extract_results(self, results: list[T]) -> list[T]:
        """Post-process search results."""
        return results

    def search(
        self,
        query: str,
        region: str = "us-en",
        safesearch: str = "moderate",
        timelimit: str | None = None,
        page: int = 1,
        **kwargs: str,
    ) -> list[T] | None:
        """Search the engine."""
        payload = self.build_payload(
            query=query, region=region, safesearch=safesearch, timelimit=timelimit, page=page, **kwargs
        )
        if self.search_method == "GET":
            html_text = self.request(self.search_method, self.search_url, params=payload, headers=self.search_headers)
        else:
            html_text = self.request(self.search_method, self.search_url, data=payload, headers=self.search_headers)
        if not html_text:
            return None
        results = self.extract_results(html_text)
        return self.post_extract_results(results)



"""Utilities."""

import re
import unicodedata
from contextlib import suppress
from datetime import datetime, timezone
from html import unescape
from urllib.parse import unquote

from .exceptions import SearchException

_REGEX_STRIP_TAGS = re.compile("<.*?>")


def _extract_vqd(html_bytes: bytes, query: str) -> str:
    """Extract vqd from html bytes."""
    for c1, c1_len, c2 in (
        (b'vqd="', 5, b'"'),
        (b"vqd=", 4, b"&"),
        (b"vqd='", 5, b"'"),
    ):
        with suppress(ValueError):
            start = html_bytes.index(c1) + c1_len
            end = html_bytes.index(c2, start)
            return html_bytes[start:end].decode()

    msg = f"_extract_vqd() {query=} Could not extract vqd."
    raise SearchException(msg)


def _normalize_url(url: str) -> str:
    """Unquote URL and replace spaces with '+'."""
    return unquote(url).replace(" ", "+") if url else ""


def _normalize_text(raw: str) -> str:
    """Normalize text.

    Strip HTML tags, unescape HTML entities, normalize Unicode,
    remove "c" category characters, and collapse whitespace.
    """
    if not raw:
        return ""

    # 1. Strip HTML tags
    text = _REGEX_STRIP_TAGS.sub("", raw)

    # 2. Unescape HTML entities
    text = unescape(text)

    # 3. Unicode normalization
    text = unicodedata.normalize("NFC", text)

    # 4. Remove "C" category characters
    c_to_none = {ord(ch): None for ch in set(text) if unicodedata.category(ch)[0] == "C"}
    if c_to_none:
        text = text.translate(c_to_none)

    # 5. Collapse whitespace
    return " ".join(text.split())


def _normalize_date(date: int | str) -> str:
    """Normalize date from integer to ISO format if applicable."""
    return datetime.fromtimestamp(date, timezone.utc).isoformat() if isinstance(date, int) else date


def _expand_proxy_tb_alias(proxy: str | None) -> str | None:
    """Expand "tb" to a full proxy URL if applicable."""
    return "socks5h://127.0.0.1:9150" if proxy == "tb" else proxy



"""Result classes."""

from abc import ABC
from collections import Counter
from collections.abc import Callable, Mapping
from dataclasses import dataclass, field
from typing import Any, ClassVar, Generic, TypeVar

from .utils import _normalize_date, _normalize_text, _normalize_url

T = TypeVar("T")


class BaseResult:
    """Base class for all results. Contains normalization functions."""

    _normalizers: ClassVar[Mapping[str, Callable[[Any], str]]] = {
        "title": _normalize_text,
        "body": _normalize_text,
        "href": _normalize_url,
        "url": _normalize_url,
        "thumbnail": _normalize_url,
        "image": _normalize_url,
        "date": _normalize_date,
        "author": _normalize_text,
        "publisher": _normalize_text,
        "info": _normalize_text,
    }

    def __setattr__(self, name: str, value: str) -> None:
        """Override setattr to apply normalization functions to certain attributes."""
        if value and (normalizer := self._normalizers.get(name)):
            value = normalizer(value)
        object.__setattr__(self, name, value)


@dataclass
class TextResult(BaseResult):
    """Text search result."""

    title: str = ""
    href: str = ""
    body: str = ""


@dataclass
class ImagesResult(BaseResult):
    """Image search result."""

    title: str = ""
    image: str = ""
    thumbnail: str = ""
    url: str = ""
    height: str = ""
    width: str = ""
    source: str = ""


@dataclass
class NewsResult(BaseResult):
    """News search result."""

    date: str = ""
    title: str = ""
    body: str = ""
    url: str = ""
    image: str = ""
    source: str = ""


@dataclass
class VideosResult(BaseResult):
    """Video search result."""

    title: str = ""
    content: str = ""
    description: str = ""
    duration: str = ""
    embed_html: str = ""
    embed_url: str = ""
    image_token: str = ""
    images: dict[str, str] = field(default_factory=dict)
    provider: str = ""
    published: str = ""
    publisher: str = ""
    statistics: dict[str, str] = field(default_factory=dict)
    uploader: str = ""


@dataclass
class BooksResult(BaseResult):
    """Book search result."""

    title: str = ""
    author: str = ""
    publisher: str = ""
    info: str = ""
    url: str = ""
    thumbnail: str = ""


class ResultsAggregator(ABC, Generic[T]):
    """Aggregates incoming results.

    Items are deduplicated by `cache_field`. Append just increments a counter;
    `extract_results` returns items sorted by descending frequency.
    """

    def __init__(self, cache_fields: set[str]) -> None:
        if not cache_fields:
            msg = "At least one cache_field must be provided"
            raise ValueError(msg)
        self.cache_fields = set(cache_fields)
        self._counter: Counter[str] = Counter()
        self._cache: dict[str, T] = {}

    def _get_key(self, item: T) -> str:
        for key in item.__dict__:
            if key in self.cache_fields:
                return str(item.__dict__[key])
        msg = f"Item {item!r} has none of the cache fields {self.cache_fields}"
        raise AttributeError(msg)

    def __len__(self) -> int:
        """Return the number of items in the cache."""
        return len(self._cache)

    def append(self, item: T) -> None:
        """Add an item to the cache.

        Register an occurrence of `item`. First time we see its key,
        we store the item; every time we bump the counter.
        """
        key = self._get_key(item)
        if key not in self._cache or len(item.__dict__.get("body", "")) > len(
            self._cache[key].__dict__.get("body", ""),
        ):
            self._cache[key] = item
        self._counter[key] += 1

    def extend(self, items: list[T]) -> None:
        """Add a list of items to the cache."""
        for item in items:
            self.append(item)

    def extract_dicts(self) -> list[dict[str, Any]]:
        """Return a list of items, sorted by descending frequency. Each item is returned as a dict."""
        return [self._cache[key].__dict__ for key, _ in self._counter.most_common()]



"""Seekify exceptions."""


class SearchException(Exception):
    """Base exception class for seekify."""


class RatelimitException(SearchException):
    """Raised for rate limit exceeded errors during API requests."""


class TimeoutException(SearchException):
    """Raised for timeout errors during API requests."""



"""Duckduckgo search engine implementation."""

from collections.abc import Mapping
from typing import Any, ClassVar, TypeVar

from fake_useragent import UserAgent

from seekify.base import BaseSearchEngine
from seekify.duckduckgo_client import DuckDuckGoClient
from seekify.results import TextResult

ua = UserAgent()

T = TypeVar("T")


class Duckduckgo(BaseSearchEngine[TextResult]):
    """Duckduckgo search engine."""

    name = "duckduckgo"
    category = "text"
    provider = "bing"

    search_url = "https://html.duckduckgo.com/html/"
    search_method = "POST"

    items_xpath = "//div[contains(@class, 'body')]"
    elements_xpath: ClassVar[Mapping[str, str]] = {"title": ".//h2//text()", "href": "./a/@href", "body": "./a//text()"}

    headers: ClassVar[dict[str, str]] = {"User-Agent": ua.random}

    def __init__(self, proxy: str | None = None, timeout: int | None = None, *, verify: bool = True) -> None:
        """Temporary, delete when HttpClient is fixed."""
        self.client = DuckDuckGoClient(headers=self.headers, proxy=proxy, timeout=timeout, verify=verify)  # type: ignore[assignment]
        self.results: list[T] = []  # type: ignore[valid-type]

    def build_payload(
        self,
        query: str,
        region: str,
        safesearch: str,  # noqa: ARG002
        timelimit: str | None,
        page: int = 1,
        **kwargs: str,  # noqa: ARG002
    ) -> dict[str, Any]:
        """Build a payload for the search request."""
        payload = {"q": query, "b": "", "l": region}
        if page > 1:
            payload["s"] = f"{10 + (page - 2) * 15}"
        if timelimit:
            payload["df"] = timelimit
        return payload

    def post_extract_results(self, results: list[TextResult]) -> list[TextResult]:
        """Post-process search results."""
        return [r for r in results if not r.href.startswith("https://duckduckgo.com/y.js?")]



"""Bing search engine implementation."""

import base64
from collections.abc import Mapping
from time import time
from typing import Any, ClassVar
from urllib.parse import parse_qs, urlparse

from seekify.base import BaseSearchEngine
from seekify.results import TextResult


def unwrap_bing_url(raw_url: str) -> str | None:
    """Decode the Bing-wrapped raw_url to extract the original url."""
    parsed = urlparse(raw_url)
    u_vals = parse_qs(parsed.query).get("u", [])
    if not u_vals:
        return None

    u = u_vals[0]
    if len(u) <= 2:
        return None

    # Drop the first two characters, pad to a multiple of 4, then decode
    b64_part = u[2:]
    padding = "=" * (-len(b64_part) % 4)
    decoded = base64.urlsafe_b64decode(b64_part + padding)
    return decoded.decode()


class Bing(BaseSearchEngine[TextResult]):
    """Bing search engine."""

    disabled = True  # !!!

    name = "bing"
    category = "text"
    provider = "bing"

    search_url = "https://www.bing.com/search"
    search_method = "GET"

    items_xpath = "//li[contains(@class, 'b_algo')]"
    elements_xpath: ClassVar[Mapping[str, str]] = {
        "title": ".//h2/a//text()",
        "href": ".//h2/a/@href",
        "body": ".//p//text()",
    }

    def build_payload(
        self,
        query: str,
        region: str,
        safesearch: str,  # noqa: ARG002
        timelimit: str | None,
        page: int = 1,
        **kwargs: str,  # noqa: ARG002
    ) -> dict[str, Any]:
        """Build a payload for the Bing search request."""
        country, lang = region.lower().split("-")
        payload = {"q": query, "pq": query, "cc": lang}
        cookies = {
            "_EDGE_CD": f"m={lang}-{country}&u={lang}-{country}",
            "_EDGE_S": f"mkt={lang}-{country}&ui={lang}-{country}",
        }
        self.client.client.set_cookies("https://www.bing.com", cookies)
        if timelimit:
            d = int(time() // 86400)
            code = f"ez5_{d - 365}_{d}" if timelimit == "y" else "ez" + {"d": "1", "w": "2", "m": "3"}[timelimit]
            payload["filters"] = f'ex1:"{code}"'
        if page > 1:
            payload["first"] = f"{(page - 1) * 10}"
            payload["FORM"] = f"PERE{page - 2 if page > 2 else ''}"
        return payload

    def post_extract_results(self, results: list[TextResult]) -> list[TextResult]:
        """Post-process search results."""
        post_results = []
        for result in results:
            if result.href.startswith("https://www.bing.com/aclick?"):
                continue
            if result.href.startswith("https://www.bing.com/ck/a?"):
                result.href = unwrap_bing_url(result.href) or result.href
            post_results.append(result)
        return post_results



"""Grokipedia text search engine."""

import json
import logging
from typing import Any

from seekify.base import BaseSearchEngine
from seekify.results import TextResult

logger = logging.getLogger(__name__)


class Grokipedia(BaseSearchEngine[TextResult]):
    """Grokipedia text search engine."""

    name = "grokipedia"
    category = "text"
    provider = "grokipedia"
    priority = 1.9

    search_url = "https://grokipedia.com/api/typeahead"
    search_method = "GET"

    def build_payload(
        self,
        query: str,
        region: str,  # noqa: ARG002
        safesearch: str,  # noqa: ARG002
        timelimit: str | None,  # noqa: ARG002
        page: int = 1,  # noqa: ARG002
        **kwargs: str,  # noqa: ARG002
    ) -> dict[str, Any]:
        """Build a payload for the search request."""
        payload: dict[str, Any] = {"query": query, "limit": "1"}
        return payload

    def extract_results(self, html_text: str) -> list[TextResult]:
        """Extract search results from html text."""
        json_data = json.loads(html_text)
        items = json_data.get("results", [])
        if not items:
            return []

        result = TextResult()
        result.title = items[0].get("title", "").strip("_")
        body = items[0].get("snippet", "")
        result.body = body.split("\n\n", 1)[1] if "\n\n" in body else body
        result.href = f"https://grokipedia.com/page/{items[0]['slug']}"
        return [result]



"""Seekify - A metasearch library for simple and distributed search.

A metasearch library that aggregates results from diverse web search services.
"""

import importlib
import logging
import threading
from typing import TYPE_CHECKING, Any, cast

__version__ = "1.0.0"
__all__ = ("Search",)


if TYPE_CHECKING:
    from .search import Search

# A do-nothing logging handler
# https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
logging.getLogger("ddgs").addHandler(logging.NullHandler())


class _ProxyMeta(type):
    _lock: threading.Lock = threading.Lock()
    _real_cls: type["Search"] | None = None

    @classmethod
    def _load_real(cls) -> type["Search"]:
        if cls._real_cls is None:
            with cls._lock:
                if cls._real_cls is None:
                    cls._real_cls = importlib.import_module(".search", package=__name__).Search
                    globals()["Search"] = cls._real_cls
        return cls._real_cls

    def __call__(cls, *args: Any, **kwargs: Any) -> "Search":  
        real = type(cls)._load_real()
        return real(*args, **kwargs)

    def __getattr__(cls, name: str) -> Any:  
        return getattr(type(cls)._load_real(), name)

    def __dir__(cls) -> list[str]:
        base = set(super().__dir__())
        loaded_names = set(dir(type(cls)._load_real()))
        return sorted(base | (loaded_names - base))


class _SearchProxy(metaclass=_ProxyMeta):
    """Proxy class for lazy-loading the real Search implementation."""


Search: type[Search] = cast("type[Search]", _SearchProxy)  # type: ignore[no-redef]



"""Bing news engine implementation."""

import re
from collections.abc import Mapping
from contextlib import suppress
from datetime import datetime, timedelta, timezone
from typing import Any, ClassVar

from seekify.base import BaseSearchEngine
from seekify.results import NewsResult

DATE_RE = re.compile(r"\b(\d+)\s*(days|tagen|jours|giorni|dias|días|дн\.|день)?\b", re.IGNORECASE)


def extract_date(pub_date_str: str) -> str:
    """Extract date from string."""
    # Try parsing the date with predefined formats
    date_formats = ["%d.%m.%Y", "%m/%d/%Y", "%d/%m/%Y"]
    for date_format in date_formats:
        with suppress(ValueError):
            return datetime.strptime(pub_date_str, date_format).astimezone(timezone.utc).isoformat()

    # Search for relative date expressions
    match = DATE_RE.search(pub_date_str)
    if match:
        days_ago = int(match.group(1))
        return (datetime.now(timezone.utc) - timedelta(days=days_ago)).replace(microsecond=0).isoformat()

    # Return the original string if no date is found
    return pub_date_str


class BingNews(BaseSearchEngine[NewsResult]):
    """Bing news engine."""

    name = "bing"
    category = "news"
    provider = "bing"

    search_url = "https://www.bing.com/news/infinitescrollajax"
    search_method = "GET"

    items_xpath = "//div[contains(@class, 'newsitem')]"
    elements_xpath: ClassVar[Mapping[str, str]] = {
        "date": ".//span[@aria-label]//@aria-label",
        "title": "@data-title",
        "body": ".//div[@class='snippet']//text()",
        "url": "@url",
        "image": ".//a[contains(@class, 'image')]//@src",
        "source": "@data-author",
    }

    def build_payload(
        self,
        query: str,
        region: str,
        safesearch: str,  # noqa: ARG002
        timelimit: str | None,
        page: int = 1,
        **kwargs: str,  # noqa: ARG002
    ) -> dict[str, Any]:
        """Build a payload for the Bing search request."""
        country, lang = region.lower().split("-")
        payload = {
            "q": query,
            "InfiniteScroll": "1",
            "first": f"{page * 10 + 1}",
            "SFX": f"{page}",
            "cc": country,
            "setlang": lang,
        }
        if timelimit:
            payload["qft"] = {
                "d": 'interval="4"',  # doesn't exist so it's the same as one hour
                "w": 'interval="7"',
                "m": 'interval="9"',
                "y": 'interval="9"',  # doesn't exist so it's the same as month
            }[timelimit]
        return payload

    def post_extract_results(self, results: list[NewsResult]) -> list[NewsResult]:
        """Post-process search results."""
        for result in results:
            result.date = extract_date(result.date)
            result.image = f"https://www.bing.com{result.image.split('&')[0]}" if result.image else ""
        return results



"""Temporary HTTP client for 'backend=duckduckgo'. Delete when HttpClient is fixed."""

import logging
import ssl
from random import SystemRandom
from types import TracebackType
from typing import TYPE_CHECKING, Any

import h2
import httpcore
import httpx

from .exceptions import SearchException, TimeoutException

if TYPE_CHECKING:
    from collections.abc import Callable


logger = logging.getLogger(__name__)
random = SystemRandom()


class Response:
    """HTTP response."""

    __slots__ = ("content", "status_code", "text")

    def __init__(self, status_code: int, content: bytes, text: str) -> None:
        self.status_code = status_code
        self.content = content
        self.text = text


class DuckDuckGoClient:
    """Specialized HTTP client for DuckDuckGo."""

    def __init__(
        self,
        headers: dict[str, str] | None = None,
        proxy: str | None = None,
        timeout: int | None = 10,
        *,
        verify: bool | str = True,
    ) -> None:
        """Initialize the DuckDuckGoClient object.

        Args:
            headers (dict, optional): headers for the HTTP client.
            proxy (str, optional): proxy for the HTTP client, supports http/https/socks5 protocols.
                example: "http://user:pass@example.com:3128". Defaults to None.
            timeout (int, optional): Timeout value for the HTTP client. Defaults to 10.
            verify: (bool | str):  True to verify, False to skip or str path to a PEM file. Defaults to True.

        """
        self.client = httpx.Client(
            headers=headers,
            proxy=proxy,
            timeout=timeout,
            verify=_get_random_ssl_context(verify=verify) if verify else False,
            follow_redirects=False,
            http2=True,
        )

    def request(self, *args: Any, **kwargs: Any) -> Response:  # noqa: ANN401
        """Make a request to the HTTP client."""
        with Patch():
            try:
                resp = self.client.request(*args, **kwargs)
                return Response(status_code=resp.status_code, content=resp.content, text=resp.text)
            except Exception as ex:
                if "timed out" in f"{ex}":
                    msg = f"Request timed out: {ex!r}"
                    raise TimeoutException(msg) from ex
                msg = f"{type(ex).__name__}: {ex!r}"
                raise SearchException(msg) from ex

    def get(self, *args: Any, **kwargs: Any) -> Response:  # noqa: ANN401
        """Make a GET request to the HTTP client."""
        return self.request(*args, method="GET", **kwargs)

    def post(self, *args: Any, **kwargs: Any) -> Response:  # noqa: ANN401
        """Make a POST request to the HTTP client."""
        return self.request(*args, method="POST", **kwargs)


# SSL
DEFAULT_CIPHERS = [  # https://developers.cloudflare.com/ssl/reference/cipher-suites/recommendations/
    "TLS_AES_128_GCM_SHA256", "TLS_AES_256_GCM_SHA384", "TLS_CHACHA20_POLY1305_SHA256",
    # Modern:
    "ECDHE-ECDSA-AES128-GCM-SHA256", "ECDHE-ECDSA-CHACHA20-POLY1305", "ECDHE-RSA-AES128-GCM-SHA256",
    "ECDHE-RSA-CHACHA20-POLY1305", "ECDHE-ECDSA-AES256-GCM-SHA384", "ECDHE-RSA-AES256-GCM-SHA384",
    # Compatible:
    "ECDHE-ECDSA-AES128-GCM-SHA256", "ECDHE-ECDSA-CHACHA20-POLY1305", "ECDHE-RSA-AES128-GCM-SHA256",
    "ECDHE-RSA-CHACHA20-POLY1305", "ECDHE-ECDSA-AES256-GCM-SHA384", "ECDHE-RSA-AES256-GCM-SHA384",
    "ECDHE-ECDSA-AES128-SHA256", "ECDHE-RSA-AES128-SHA256", "ECDHE-ECDSA-AES256-SHA384",  "ECDHE-RSA-AES256-SHA384",
    # Legacy:
    "ECDHE-ECDSA-AES128-SHA", "ECDHE-RSA-AES128-SHA", "AES128-GCM-SHA256", "AES128-SHA256", "AES128-SHA",
    "ECDHE-RSA-AES256-SHA", "AES256-GCM-SHA384", "AES256-SHA256", "AES256-SHA", "DES-CBC3-SHA",
]  # fmt: skip


def _get_random_ssl_context(*, verify: bool | str) -> ssl.SSLContext:
    ssl_context = ssl.create_default_context(cafile=verify if isinstance(verify, str) else None)
    shuffled_ciphers = random.sample(DEFAULT_CIPHERS[9:], len(DEFAULT_CIPHERS) - 9)
    ssl_context.set_ciphers(":".join(DEFAULT_CIPHERS[:9] + shuffled_ciphers))
    commands: list[None | Callable[[ssl.SSLContext], None]] = [
        None,
        lambda context: setattr(context, "maximum_version", ssl.TLSVersion.TLSv1_2),
        lambda context: setattr(context, "minimum_version", ssl.TLSVersion.TLSv1_3),
        lambda context: setattr(context, "options", context.options | ssl.OP_NO_TICKET),
    ]
    random_command = random.choice(commands)
    if random_command:
        random_command(ssl_context)
    return ssl_context


class Patch:
    """Patch the HTTP2Connection._send_connection_init method."""

    def __enter__(self) -> None:
        """Enter the context manager."""

        def _send_connection_init(self: httpcore._sync.http2.HTTP2Connection, request: httpcore.Request) -> None:
            """
            Custom connection initialization to mimic different browser fingerprints.
            
            This overrides the default HTTP/2 settings to use randomized values for:
            - Window size
            - Header table size
            - Max frame size
            - Max concurrent streams
            
            This randomization helps in avoiding bot detection systems that fingerprint
            standard HTTP client libraries.
            """
            self._h2_state.local_settings = h2.settings.Settings(
                client=True,
                initial_values={
                    h2.settings.SettingCodes.INITIAL_WINDOW_SIZE: random.randint(100, 200),
                    h2.settings.SettingCodes.HEADER_TABLE_SIZE: random.randint(4000, 5000),
                    h2.settings.SettingCodes.MAX_FRAME_SIZE: random.randint(16384, 65535),
                    h2.settings.SettingCodes.MAX_CONCURRENT_STREAMS: random.randint(100, 200),
                    h2.settings.SettingCodes.MAX_HEADER_LIST_SIZE: random.randint(65500, 66500),
                    h2.settings.SettingCodes.ENABLE_CONNECT_PROTOCOL: random.randint(0, 1),
                    h2.settings.SettingCodes.ENABLE_PUSH: random.randint(0, 1),
                },
            )
            self._h2_state.initiate_connection()
            self._h2_state.increment_flow_control_window(2**24)
            self._write_outgoing_data(request)

        # Monkey-patch the method on the httpcore class
        self.original_send_connection_init = httpcore._sync.http2.HTTP2Connection._send_connection_init
        httpcore._sync.http2.HTTP2Connection._send_connection_init = _send_connection_init  # type: ignore[method-assign]

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        """Exit the context manager."""
        httpcore._sync.http2.HTTP2Connection._send_connection_init = self.original_send_connection_init  # type: ignore[method-assign]



import subprocess
import sys
import pytest

def run_cli_command(args):
    """Run CLI command and return stdout, stderr, and return code."""
    cmd = [sys.executable, "-m", "seekify.cli"] + args
    result = subprocess.run(cmd, capture_output=True, text=True)
    return result

def test_cli_version():
    result = run_cli_command(["version"])
    assert result.returncode == 0
    assert "1.0.0" in result.stdout

def test_cli_text_search():
    # Test basic text search with a specific backend
    result = run_cli_command(["text", "-q", "python", "-m", "1", "-b", "duckduckgo"])
    if result.returncode != 0:
        print(f"CLI Output: {result.stdout}")
        print(f"CLI Error: {result.stderr}")
    assert result.returncode == 0
    # Check for likely content
    assert "python" in result.stdout.lower() or "programming" in result.stdout.lower()

def test_cli_json_output(tmp_path):
    # Test output to JSON
    output_file = tmp_path / "results.json"
    result = run_cli_command(["text", "-q", "test", "-m", "1", "-o", str(output_file)])
    assert result.returncode == 0
    assert output_file.exists()



import logging
import pytest
from seekify import Search

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Engines to test
TEXT_ENGINES = [
    "bing", "brave", "duckduckgo", "google", 
    "grokipedia", "mojeek", "yandex", "yahoo", "wikipedia"
]
IMAGE_ENGINES = ["duckduckgo"]
VIDEO_ENGINES = ["duckduckgo"]
NEWS_ENGINES = ["bing", "duckduckgo", "yahoo"]
BOOK_ENGINES = ["annasarchive"]

@pytest.mark.parametrize("engine", TEXT_ENGINES)
def test_text_search(engine):
    logger.info(f"Testing text search with backend: {engine}")
    with Search() as s:
        results = s.text("python programming", backend=engine, max_results=3)
        assert results is not None
        # Some engines might legitimately return 0 results or be blocked/rate-limited. 
        # We warn instead of failing strictly if empty, but we expect a list.
        if not results:
            logger.warning(f"No results returned for {engine}")
        else:
            assert len(results) > 0
            assert "title" in results[0]
            assert "href" in results[0]

@pytest.mark.parametrize("engine", IMAGE_ENGINES)
def test_image_search(engine):
    logger.info(f"Testing image search with backend: {engine}")
    with Search() as s:
        results = s.images("cute cat", backend=engine, max_results=3)
        assert results is not None
        if not results:
            logger.warning(f"No results returned for {engine}")
        else:
            assert len(results) > 0
            assert "image" in results[0]

@pytest.mark.parametrize("engine", VIDEO_ENGINES)
def test_video_search(engine):
    logger.info(f"Testing video search with backend: {engine}")
    with Search() as s:
        results = s.videos("python tutorial", backend=engine, max_results=3)
        assert results is not None
        if not results:
            logger.warning(f"No results returned for {engine}")
        else:
            assert len(results) > 0
            assert "content" in results[0]

@pytest.mark.parametrize("engine", NEWS_ENGINES)
def test_news_search(engine):
    logger.info(f"Testing news search with backend: {engine}")
    try:
        with Search() as s:
            results = s.news("technology", backend=engine, max_results=3)
            assert results is not None
            if not results:
                logger.warning(f"No results returned for {engine}")
            else:
                assert len(results) > 0
                assert "title" in results[0]
    except Exception as e:
        logger.warning(f"News search failed for {engine}: {e}")
        # Mark as skip or pass with warning to avoid failing the whole suite for one flaky provider
        pytest.skip(f"Provider {engine} failed: {e}")

@pytest.mark.parametrize("engine", BOOK_ENGINES)
def test_book_search(engine):
    logger.info(f"Testing book search with backend: {engine}")
    try:
        with Search() as s:
            results = s.books("python", backend=engine, max_results=3)
            assert results is not None
            if not results:
                logger.warning(f"No results returned for {engine}")
            else:
                assert len(results) > 0
    except Exception as e:
        # Anna's archive can be unstable or blocked
        logger.warning(f"Book search failed for {engine}: {e}")



"""Yahoo! News search engine."""

import logging
import re
from collections.abc import Callable, Mapping
from datetime import datetime, timedelta, timezone
from typing import Any, ClassVar
from urllib.parse import unquote_plus

from seekify.base import BaseSearchEngine
from seekify.results import NewsResult

logger = logging.getLogger(__name__)

DATE_RE = re.compile(r"\b(\d+)\s*(year|month|week|day|hour|minute)s?\b", re.IGNORECASE)
DATE_UNITS: dict[str, Callable[[int], timedelta]] = {
    "minute": lambda n: timedelta(minutes=n),
    "hour": lambda n: timedelta(hours=n),
    "day": lambda n: timedelta(days=n),
    "week": lambda n: timedelta(weeks=n),
    "month": lambda n: timedelta(days=30 * n),
    "year": lambda n: timedelta(days=365 * n),
}


def extract_date(pub_date_str: str) -> str:
    """Extract date from string."""
    now = datetime.now(timezone.utc)
    m = DATE_RE.search(pub_date_str)
    if not m:
        return pub_date_str

    number = int(m.group(1))
    unit = m.group(2).lower()
    delta = DATE_UNITS[unit](number)
    dt = (now - delta).replace(microsecond=0)
    return dt.isoformat()


def extract_url(u: str) -> str:
    """Sanitize url."""
    url = u.split("/RU=", 1)[1].split("/RK=", 1)[0].split("?", 1)[0]
    return unquote_plus(url)


def extract_image(u: str) -> str:
    """Sanitize image url."""
    idx = u.find("-/")
    return u[idx + 2 :] if idx != -1 else u


def extract_source(s: str) -> str:
    """Remove ' via Yahoo' from string."""
    return s.split(" ·  via Yahoo")[0]


class YahooNews(BaseSearchEngine[NewsResult]):
    """Yahoo news search engine."""

    name = "yahoo"
    category = "news"
    provider = "yahoo"

    search_url = "https://news.search.yahoo.com/search"
    search_method = "GET"

    items_xpath = "//div[@id='web']//li[a]"
    elements_xpath: ClassVar[Mapping[str, str]] = {
        "date": ".//span[contains(@class, 'time')]//text()",
        "title": ".//h4//text()",
        "body": ".//p//text()",
        "url": ".//h4/a/@href",
        "image": "(.//img/@data-src | .//img/@src)[1]",
        "source": ".//span[contains(@class, 'source')]//text()",
    }

    def build_payload(
        self,
        query: str,
        region: str,  # noqa: ARG002
        safesearch: str,  # noqa: ARG002
        timelimit: str | None,
        page: int = 1,
        **kwargs: str,  # noqa: ARG002
    ) -> dict[str, Any]:
        """Build a payload for the search request."""
        payload = {"p": query}
        if page > 1:
            payload["b"] = f"{(page - 1) * 10 + 1}"
        if timelimit:
            payload["btf"] = timelimit
        return payload

    def post_extract_results(self, results: list[NewsResult]) -> list[NewsResult]:
        """Post-process search results."""
        try:
            for result in results:
                result.date = extract_date(result.date)
                result.url = extract_url(result.url)
                result.image = extract_image(result.image)
                result.source = extract_source(result.source)
        except Exception as ex:  # noqa: BLE001
            logger.warning("Error post-processing results: %r", ex)
        return results



"""Google search engine implementation."""

from collections.abc import Mapping
from random import SystemRandom
from typing import Any, ClassVar

from seekify.base import BaseSearchEngine
from seekify.results import TextResult

random = SystemRandom()


def get_ua() -> str:
    """Return one random User-Agent string."""
    patterns = [
        "Opera/9.80 (J2ME/MIDP; Opera Mini/{v}/{b}; U; {l}) Presto/{p} Version/{f}",
        "Opera/9.80 (Android; Linux; Opera Mobi/{b}; U; {l}) Presto/{p} Version/{f}",
        "Opera/9.80 (iPhone; Opera Mini/{v}/{b}; U; {l}) Presto/{p} Version/{f}",
        "Opera/9.80 (iPad; Opera Mini/{v}/{b}; U; {l}) Presto/{p} Version/{f}",
    ]
    mini_versions = ["4.0", "5.0.17381", "7.1.32444", "9.80"]
    mobi_builds = ["27", "447", "ADR-1011151731"]
    builds = ["18.678", "24.743", "503"]
    prestos = ["2.6.35", "2.7.60", "2.8.119"]
    finals = ["10.00", "11.10", "12.16"]
    langs = ["en-US", "en-GB", "de-DE", "fr-FR", "es-ES", "ru-RU", "zh-CN"]
    fallback = "Opera/9.80 (iPad; Opera Mini/5.0.17381/503; U; eu) Presto/2.6.35 Version/11.10"

    try:
        p = random.choice(patterns)
        vals = {
            "l": random.choice(langs),
            "p": random.choice(prestos),
            "f": random.choice(finals),
        }
        if "{v}" in p:
            vals["v"] = random.choice(mini_versions)
        if "{b}" in p:
            vals["b"] = random.choice(mobi_builds) if "Opera Mobi" in p else random.choice(builds)
        return p.format(**vals)
    except Exception:  # noqa: BLE001
        return fallback


class Google(BaseSearchEngine[TextResult]):
    """Google search engine."""

    name = "google"
    category = "text"
    provider = "google"

    search_url = "https://www.google.com/search"
    search_method = "GET"
    search_headers: ClassVar[dict[str, str]] = {"User-Agent": get_ua()}

    items_xpath = "//div[div[@data-hveid]//div[h3]]"
    elements_xpath: ClassVar[Mapping[str, str]] = {
        "title": ".//h3//text()",
        "href": ".//a/@href",
        "body": "./div/div/div[2]//text()",
    }

    def build_payload(
        self,
        query: str,
        region: str,
        safesearch: str,
        timelimit: str | None,
        page: int = 1,
        **kwargs: str,  # noqa: ARG002
    ) -> dict[str, Any]:
        """Build a payload for the Google search request."""
        safesearch_base = {"on": "2", "moderate": "1", "off": "0"}
        start = (page - 1) * 10
        payload = {
            "q": query,
            "filter": safesearch_base[safesearch.lower()],
            "start": str(start),
        }
        country, lang = region.split("-")
        payload["hl"] = f"{lang}-{country.upper()}"  # interface language
        payload["lr"] = f"lang_{lang}"  # restricts to results written in a particular language
        payload["cr"] = f"country{country.upper()}"  # restricts to results written in a particular country
        if timelimit:
            payload["tbs"] = f"qdr:{timelimit}"
        return payload

    def post_extract_results(self, results: list[TextResult]) -> list[TextResult]:
        """Post-process search results."""
        post_results = []
        for result in results:
            if result.href.startswith("/url?q="):
                result.href = result.href.split("?q=")[1].split("&")[0]
            post_results.append(result)
        return post_results
